(*  Title:      Config.thy
    ID:         $Id:$
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Configurations, i.e. parallel activities for a functional
                ASP, based on the sigma-calculus and basic activities 
                defined in ASP.thy
                Def of configurations and reductions
      cf. Well-Formedness.thy for proof of well-formedness of the reduction presented here
   Name convention: 
               m??? = futmap == mapping fut-> term
               f??? = future id
               a, b = terms
               c??? = config
               A, B, G = activity
               E = contexts
*)

header {* Configurations for ASP *}

theory Config imports ContextLocalRed Infinite_Set begin

declare cCompose_and_Fill[simp]

(* noFV returns true it t has no free variable;
 for defining it we define FVtest t n ehre n is a nat giving the number of 
binders we have entered *)
consts FVtestVar :: "Variable \<Rightarrow> nat \<Rightarrow> bool"
       FVtest :: "dB \<Rightarrow> nat \<Rightarrow> bool"
       optionFVtest :: "dB option \<Rightarrow> nat \<Rightarrow> bool"

primrec
  FVSelf:    "FVtestVar (Self i) n = (i < n)"
  FVParam:    "FVtestVar (Param i) n = (i < n)"

primrec 
  FVVar:    "FVtest (Var x) n = FVtestVar x n"
  FVActive: "FVtest (Active a) n = FVtest a n" 
  FVActRef: "FVtest (ActRef r) n = True"
  FVFutRef: "FVtest (FutRef f) n = True"
  FVCall:   "FVtest (Call a l b) n = ((FVtest a n) \<and> (FVtest b n))"
  FVUpd:    "FVtest (Upd a l b) n = ((FVtest a n) & (FVtest b (n+1)))"
  FVObj:    "FVtest (Obj f T) n = fold (op &) (\<lambda>x. optionFVtest (f x) (n+1)) True (dom f) " 
  FVoption:   "optionFVtest None n = True"
  FVoption':  "optionFVtest (Some a) n = (FVtest a n)"

constdefs noFV :: "dB \<Rightarrow> bool"   "noFV t == FVtest t 0" 

(* noRef is true for a term containing no future and no active object reference, 
   i.e. for a source term *)
consts 
  noRef :: "dB \<Rightarrow> bool"   
  optionnoRef :: "dB option \<Rightarrow> bool"   

primrec 
  norefVar:    "noRef (Var x)  = True"
  noRefActive: "noRef (Active a)  = noRef a" 
  noRefActRef: "noRef (ActRef r)  = False"
  noRefFutRef: "noRef (FutRef f)  = False"
  noRefCall:   "noRef (Call a l b)  = ((noRef a) \<and> (noRef b))"
  noRefUpd:    "noRef (Upd a l b)  = ((noRef a) & (noRef b))"
  noRefObj:    "noRef (Obj f T)  = fold (op &) (\<lambda>x. optionnoRef (f x)) True (dom f) " 
  noRefoption:   "optionnoRef None  = True"
  noRefoption':  "optionnoRef (Some a) = (noRef a)"


types futmap = "FutureRef ~=> dB"
      configuration = "ActivityRef ~=> (futmap * dB)"

subsection{*Configuration Accessors*}

(* we define operators that access the active object, the request queue 
or a single request of a given activity *)
constdefs ActiveObject :: "[configuration, ActivityRef] \<Rightarrow> dB" (infixl ".\<^sub>A" 50)
  " (ActiveObject c A)==(snd (the (c A)))"
constdefs RequestQueue ::  "[configuration, ActivityRef] \<Rightarrow> futmap" (infixl ".\<^sub>R" 50)
  " (RequestQueue c A)==(fst (the (c A)))"
constdefs OneRequest ::  "[configuration, ActivityRef, FutureRef] \<Rightarrow> dB" ("_.\<^sub>F_<_>" [50,0,0] 50)
  " (OneRequest c A fi)==(the ((fst (the (c A))) fi))"

lemma ActiveObjectE: "\<lbrakk>C.\<^sub>AA = a;A\<in>dom C\<rbrakk> \<Longrightarrow> \<exists>m. C A = Some(m,a)"
apply (unfold ActiveObject_def);
by auto;

lemma RequestQueueE: "\<lbrakk>C.\<^sub>RA = m;A\<in>dom C\<rbrakk> \<Longrightarrow> \<exists>a. C A = Some(m,a)"
apply (unfold RequestQueue_def);
by auto;

lemma OneRequestE: "\<lbrakk>C.\<^sub>FA<f> = s;A\<in>dom C;f\<in>dom(C.\<^sub>RA)\<rbrakk> \<Longrightarrow> \<exists>m a. C A = Some(m,a)\<and>m f = Some s"
apply (unfold OneRequest_def);
apply (unfold RequestQueue_def);
by auto;

lemma ActiveObjectI: " C A = Some(m,a) \<Longrightarrow> C.\<^sub>AA = a"
apply (unfold ActiveObject_def);
by auto;

lemma RequestQueueI: "C A = Some(m,a)\<Longrightarrow>C.\<^sub>RA = m"
apply (unfold RequestQueue_def);
by auto;

lemma OneRequestI: "\<lbrakk>C A = Some(m,a);m f = Some s\<rbrakk> \<Longrightarrow>C.\<^sub>FA<f> = s"
apply (unfold OneRequest_def);
by auto;

lemma ActivityDef: "A \<in> dom c \<Longrightarrow> c A = Some (c.\<^sub>RA,c.\<^sub>AA)"
by (auto simp: ActiveObject_def RequestQueue_def)

subsection{*Reduction*}
(* now the semantics, see ASPfun papers for details *)
inductive zeta :: "[configuration, configuration] => bool"  (infixl "\<rightarrow>\<^sub>\<zeta>" 50)
  where
    zeta_local [simp, intro!]: "\<lbrakk> a \<rightarrow>\<^sub>\<beta> a'; c A = Some(m,a''); m(fk) = Some a \<rbrakk> 
                           \<Longrightarrow> c\<rightarrow>\<^sub>\<zeta>  c (A \<mapsto> (m(fk \<mapsto> a'), a''))"  
|
    active [simp, intro!]: "\<lbrakk> c G = None; c A = Some(m, a''); 
                          m(fk) = Some((E::general_context)\<up>(Active a)); noFV a \<rbrakk> 
                       \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c(A \<mapsto> (m (fk \<mapsto> (E\<up>(ActRef G))), a''))(G \<mapsto> (empty, a))"  
|
    request [simp,intro]: "\<lbrakk> \<forall> A\<in> dom c. fn \<notin> dom(fst(the(c A))); 
                             c A' = Some(m',a'); m'(fk) = Some((E::general_context)\<up>(Call (ActRef B) li p));
                             c B = Some(mb, a);A'\<noteq>B; noFV p \<rbrakk> \<Longrightarrow>
                             c \<rightarrow>\<^sub>\<zeta> c (A'\<mapsto> (m' (fk \<mapsto> (E\<up>(FutRef(fn)))), a'))
                                     (B \<mapsto> (mb (fn \<mapsto> (Call a li p)), a))"
|                             
   request_self[simp,intro]: "\<lbrakk> \<forall> A\<in> dom c. fn \<notin> dom(fst(the(c A))); 
                             c A' = Some(m',a); 
                             m'(fk) = Some((E::general_context)\<up>(Call(ActRef A') li p)); noFV p\<rbrakk> \<Longrightarrow>
                          c \<rightarrow>\<^sub>\<zeta> c (A'\<mapsto> (m' (fk \<mapsto> (E\<up>(FutRef(fn))))(fn \<mapsto> (Call a li p)), a))"
| 
    update_AO [simp,intro]: "\<lbrakk> c G = None;
                       c A' = Some(m',a'); m'(fk) = Some((E::general_context)\<up>(Upd (ActRef B) li b));
                             c B = Some(mb, a);noFV (Upd (ActRef B) li b)\<rbrakk> \<Longrightarrow>
                             c \<rightarrow>\<^sub>\<zeta> c (A'\<mapsto> (m' (fk \<mapsto> (E\<up>(ActRef G))), a'))
                                     (G \<mapsto> (empty, Upd a li  b))"
|   
   reply   [simp, intro!]: "\<lbrakk> c A = Some (ma, a); ma(fk) = Some((E::general_context)\<up>  (FutRef(fi)));
                              c B = Some(mb, b); mb(fi) = Some(ai)\<rbrakk> \<Longrightarrow> 
                             c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> (ma (fk \<mapsto> (E\<up> ai)), a))"
abbreviation
  beta_reds :: "[configuration, configuration] => bool"  (infixl "~->>" 50) where
  "s ~->> t == zeta^** s t"

notation (latex)
  beta_reds  (infixl "\<rightarrow>\<^sub>\<zeta>\<^sup>*" 50)

(* a conf is finite if it has a finite number of activities and requests *)
constdefs finite_conf :: "configuration \<Rightarrow> bool"
"finite_conf c == finite (dom c) & (! A: dom c. finite (dom (fst (the(c A)))))"

lemma finite_step [rule_format]: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> finite_conf c \<longrightarrow> finite_conf c'";
apply (erule zeta.induct);
apply (unfold finite_conf_def);
apply auto;
apply force+;
done;

lemma finite_conf_rtrancl[rule_format]: "\<lbrakk> x\<rightarrow>\<^sub>\<zeta>\<^sup>*  y \<rbrakk> \<Longrightarrow> finite_conf x \<longrightarrow> finite_conf y";
apply (erule rtranclp_induct);
apply (auto simp: finite_step);
done;

(* if a set is finite, but its base type is infinite, then there is always a fresh
   element *)
lemma finite_set_subset_infinite_fresh: 
   "finite (A :: (('a :: inftype)set)) \<Longrightarrow> \<exists> x :: 'a. x \<notin> A";

apply (subgoal_tac "~finite(UNIV :: 'a set)");
apply (subgoal_tac "A <= (UNIV :: 'a set)");
apply (subgoal_tac "A ~= UNIV");
apply (blast+);
apply (rule FMap.inftype_class.axioms);
done;

(* We need to state that the domains of configurations are over infinite types.
   However, the types FutureRef and ActivityRef are already defined as datatypes
   embedding nat. Hence we can _derive_ the infinite of these types from there. *)

instance nat :: inftype
apply intro_classes;
apply (rule Finite_Set.infinite_UNIV_nat);
done;

lemma surj_fref: "surj fref"; 
apply (unfold "surj_def");
apply auto;
apply (case_tac y);
apply (rule_tac x = nat in exI);
apply assumption;
done;

lemma inj_fref: "inj fref";
apply (unfold inj_on_def);
apply simp;
done;

instance FutureRef :: inftype 
apply intro_classes;
apply (subgoal_tac "range fref = UNIV");
prefer 2;
apply (rule surj_range);
apply (rule surj_fref);
apply (erule subst);
apply (rule range_inj_infinite);
by (rule inj_fref);

(* Here is now the old axiom fresh_futref, but with the assumption that
   the set A (in the application of the lemma to be replaced by dom c)
   is finite *)
lemma fresh_futref_finite: "finite F \<Longrightarrow> \<exists>(id::FutureRef). id \<notin> F";
by (erule finite_set_subset_infinite_fresh);


(* Same procedure for ActivityRefs *)
lemma surj_aref: "surj aref"; 
apply (unfold "surj_def");
apply auto;
apply (case_tac y);
apply (rule_tac x = nat in exI);
apply assumption;
done;

lemma inj_aref: "inj aref";
apply (unfold inj_on_def);
apply simp;
done;

instance ActivityRef :: inftype 
apply intro_classes;
apply (subgoal_tac "range aref = UNIV");
prefer 2;
apply (rule surj_range);
apply (rule surj_aref);
apply (erule subst);
apply (rule range_inj_infinite);
by (rule inj_aref);

lemma fresh_actref_finite: "finite A \<Longrightarrow> \<exists>(id::ActivityRef). id \<notin> A";
by (erule finite_set_subset_infinite_fresh);


(* Here could come lemmata to infer finite(dom c) and
   (! A: dom c. finite(dom (fst c A)) from finite_conf.
   These are just normal elimination rules. *)
lemma finite_confE1: "finite_conf c \<Longrightarrow> finite (dom c)";
by (simp add: finite_conf_def);

lemma finite_confE2: "\<lbrakk> finite_conf c; c A = Some(m,a) \<rbrakk> \<Longrightarrow> finite (dom m)";
apply (simp add: finite_conf_def);
apply force;
done;


(* 2. All reachable configs are finite: 
      We first define an initial config c0 (ie a single act only consisting in a single request, 
      mapped to a tern without future or activity ref) 
      c0 is provably finite. Using finite_conf_rtrancl we get that any c with c0 ~->> c 
      is finite. So assuming reachable configs we get our old freshness axioms with
      the help of fresh_fut/actref_finite. *)

constdefs 
  init_config :: "dB \<Rightarrow> configuration"
    "init_config a == [aref 0 \<mapsto> ([fref 0 \<mapsto> a], Obj empty (Object empty))]"

(* more structure? 
  init_configs :: "configuration set"
    "init_configs == { c. \<exists> a. (noRef a) \<and> c = init_config a}"
*)

  reachable_configs :: " configuration set" 
    "reachable_configs == { c. \<exists> a. (noRef a) \<and> init_config a ~->> c}"

lemma finite_init_config: "finite_conf (init_config a)";
apply (unfold finite_conf_def init_config_def);
by auto;

lemma finite_reachable_configs: " \<forall> c \<in> reachable_configs. finite_conf c";
apply (rule ballI);
apply (unfold reachable_configs_def);
apply simp;
apply (erule exE,erule conjE);
apply (erule finite_conf_rtrancl);
by (rule finite_init_config);

(* A simpler form that is needed *)
lemma finite_reachable_configsL: "c \<in> reachable_configs \<Longrightarrow> finite (dom c)";
apply (rule finite_confE1);
apply (simp add: finite_reachable_configs);
done;

(* And one for the union of the request queues *)
lemma finite_conf_unionE: "finite_conf conf \<Longrightarrow> finite (\<Union>{dom (fst (the (conf A))) |A. A \<in> dom conf})";
apply (rule_tac B = "UN A: dom conf. dom (fst (the (conf A)))" in finite_subset);
apply force;
apply (rule finite_UN_I);
apply (erule finite_confE1);
apply (auto intro: finite_confE2);
done;

(*  From these we can then infer  the original axioms
axioms fresh_futref: " \<exists>(id::FutureRef). id \<notin>F"
       fresh_actref: " \<exists>(id::ActivityRef). id \<notin>A"

but now relative to the assumption that the configurations are reachable.
(See TypedASP)
*)

end
